@@ -20,10 +20,13 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
+static char rcsid[] ="$Id: mkisofs.c,v 1.9 1997/04/10 02:45:09 eric Rel $";
+
 /* ADD_FILES changes made by Ross Biro biro@yggdrasil.com 2/23/95 */
 
 #include <errno.h>
 #include "mkisofs.h"
+#include "config.h"
 
 #ifdef linux
 #include <getopt.h>
@@ -43,8 +46,10 @@
 #include <sys/stat.h>
 
 #ifndef VMS
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#endif
 
 #include "exclude.h"
 
@@ -55,21 +60,22 @@
 
 struct directory * root = NULL;
 
-static char version_string[] = "mkisofs v1.05";
+static char version_string[] = "mkisofs v1.11";
 
 FILE * discimage;
 unsigned int next_extent = 0;
 unsigned int last_extent = 0;
+unsigned int session_start = 0;
 unsigned int path_table_size = 0;
 unsigned int path_table[4] = {0,};
 unsigned int path_blocks = 0;
 struct iso_directory_record root_record;
-static int timezone_offset;
 char * extension_record = NULL;
 int extension_record_extent = 0;
 static  int extension_record_size = 0;
 
 /* These variables are associated with command line options */
+int use_eltorito = 0;
 int use_RockRidge = 0;
 int verbose = 0;
 int all_files  = 0;
@@ -85,6 +91,8 @@ char * abstract = ABSTRACT_DEFAULT;
 char * volset_id = VOLSET_ID_DEFAULT;
 char * volume_id = VOLUME_ID_DEFAULT;
 char * system_id = SYSTEM_ID_DEFAULT;
+char * boot_catalog = BOOT_CATALOG_DEFAULT;
+char * boot_image = BOOT_IMAGE_DEFAULT;
 
 int omit_period = 0;             /* Violates iso9660, but these are a pain */
 int transparent_compression = 0; /* So far only works with linux */
@@ -112,7 +120,7 @@ struct rcopts rcopt[] = {
   {NULL, NULL}
 };
 
-#ifdef ultrix
+#if defined(ultrix) || defined(_AUX_SOURCE)
 char *strdup(s)
 char *s;{char *c;if(c=(char *)malloc(strlen(s)+1))strcpy(c,s);return c;}
 #endif
@@ -132,6 +140,18 @@ void FDECL1(read_rcfile, char *, appname)
   if (!rcfile && errno != ENOENT)
     perror(filename);
 
+  if (!rcfile)
+    {
+      pnt = getenv("MKISOFSRC");
+      if (pnt && strlen(pnt) <= sizeof(filename))
+	{
+	  strcpy(filename, pnt);
+	  rcfile = fopen(filename, "r");
+	  if (!rcfile && errno != ENOENT)
+	    perror(filename);
+	}
+    }
+
   if (!rcfile)
     {
       pnt = getenv("HOME");
@@ -202,9 +222,13 @@ void FDECL1(read_rcfile, char *, appname)
       pnt++; /* Skip past '=' sign */
       while (*pnt == ' ' || *pnt == '\t')
 	pnt++;
-      +       /* now it is safe to NUL terminate the name */
+
+      /* now it is safe to NUL terminate the name */
+
 	*name_end = 0;
+
       /* Now get rid of trailing newline */
+
       pnt1 = pnt;
       while (*pnt1)
 	{
@@ -242,36 +266,26 @@ void usage(){
 	fprintf(stderr,"Usage:\n");
 	fprintf(stderr,
 "mkisofs [-o outfile] [-R] [-V volid] [-v] [-a] \
-[-T]\n [-l] [-d] [-V] [-D] [-L] [-p preparer] \
-[-P publisher] [ -A app_id ] [-z] \
+[-T]\n [-l] [-d] [-V] [-D] [-L] [-p preparer]"
+#ifdef ADD_FILES
+"[-i file] \n"
+#endif
+"[-P publisher] [ -A app_id ] [-z] \n \
+[-b boot_image_name] [-c boot_catalog-name] \
 [-x path -x path ...] path\n");
 	exit(1);
 }
 
-int get_iso9660_timezone_offset(){
-  struct tm gm;
-  struct tm * pt;
-  time_t ctime;
-  int local_min, gmt_min;
 
-  time(&ctime);
-  pt = gmtime(&ctime);
-  gm = *pt;
-  pt = localtime(&ctime);
-
-  if(gm.tm_year < pt->tm_year)
-    gm.tm_yday = -1;
-
-  if(gm.tm_year > pt->tm_year)
-    pt->tm_yday = -1;
-
-  gmt_min = gm.tm_min + 60*(gm.tm_hour + 24*gm.tm_yday);
-  local_min = pt->tm_min + 60*(pt->tm_hour + 24*pt->tm_yday);
-  return (gmt_min - local_min)/15;
-}
-
-
-/* Fill in date in the iso9660 format */
+/* 
+ * Fill in date in the iso9660 format 
+ *
+ * The standards  state that the timezone offset is in multiples of 15
+ * minutes, and is what you add to GMT to get the localtime.  The U.S.
+ * is always at a negative offset, from -5h to -8h (can vary a little
+ * with DST,  I guess).  The Linux iso9660 filesystem has had the sign
+ * of this wrong for ages (mkisofs had it wrong too for the longest time).
+ */
 int FDECL2(iso9660_date,char *, result, time_t, ctime){
   struct tm *local;
   local = localtime(&ctime);
@@ -281,272 +295,42 @@ int FDECL2(iso9660_date,char *, result, time_t, ctime){
   result[3] = local->tm_hour;
   result[4] = local->tm_min;
   result[5] = local->tm_sec;
-  result[6] = timezone_offset;
+
+  /* 
+   * Must recalculate proper timezone offset each time,
+   * as some files use daylight savings time and some don't... 
+   */
+  result[6] = local->tm_yday;	/* save yday 'cause gmtime zaps it */
+  local = gmtime(&ctime);
+  local->tm_year -= result[0];
+  local->tm_yday -= result[6];
+  local->tm_hour -= result[3];
+  local->tm_min -= result[4];
+  if (local->tm_year < 0) 
+    {
+      local->tm_yday = -1;
+    }
+  else 
+    {
+      if (local->tm_year > 0) local->tm_yday = 1;
+    }
+
+  result[6] = -(local->tm_min + 60*(local->tm_hour + 24*local->tm_yday)) / 15;
+
   return 0;
 }
 
-int FDECL3(iso9660_file_length,const char*, name, struct directory_entry *, sresult, 
-			int, dirflag){
-  int seen_dot = 0;
-  int seen_semic = 0;
-  char * result;
-  int priority = 32767;
-  int tildes = 0;
-  int ignore = 0;
-  int extra = 0;
-  int current_length = 0;
-  int chars_after_dot = 0;
-  int chars_before_dot = 0;
-  const char * pnt;
-  result = sresult->isorec.name;
 
-  if(strcmp(name,".") == 0){
-    if(result) *result = 0;
-    return 1;
-  };
-
-  if(strcmp(name,"..") == 0){
-    if(result) {
-	    *result++ = 1;
-	    *result++ = 0;
-    }
-    return 1;
-  };
-
-  pnt = name;
-  while(*pnt){
-#ifdef VMS
-    if(strcmp(pnt,".DIR;1") == 0) break;
-#endif
-    if(*pnt == '#') {priority = 1; pnt++; continue; };
-    if(*pnt == '~') {priority = 1; tildes++; pnt++; continue;};
-    if(*pnt == ';') {seen_semic = 1; *result++ = *pnt++; continue; };
-    if(ignore) {pnt++; continue;};
-    if(seen_semic){
-      if(*pnt >= '0' && *pnt <= '9') *result++ = *pnt;
-      extra++;
-      pnt++;
-      continue;
-    };
-    if(full_iso9660_filenames) {
-      /* Here we allow a more relaxed syntax. */
-      if(*pnt == '.') {
-	if (seen_dot) {ignore++; continue;}
-	seen_dot++;
-      }
-      if(current_length < 30) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
-    } else { /* Dos style filenames */
-      if(*pnt == '.') {
-        if (!chars_before_dot && !allow_leading_dots) {
-	  /* DOS can't read files with dot first */
-          chars_before_dot++;
-          if (result) *result++ = '_'; /* Substitute underscore */
-        } else {
-          if (seen_dot) {ignore++; continue;}
-	  if(result) *result++ = '.';
-	  seen_dot++;
-        }
-      } else if (seen_dot) {
-	if(chars_after_dot < 3) {
-	  chars_after_dot++;
-	  if(result) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
-	}
-      } else {
-	if(chars_before_dot < 8) {
-	  chars_before_dot++;
-	  if(result) *result++ = (islower(*pnt) ? toupper(*pnt) : *pnt);
-	};
-      };
-    };
-    current_length++;
-    pnt++;
-  };
-  
-  if(tildes == 2){
-    int prio1 = 0;
-    pnt = name;
-    while (*pnt && *pnt != '~') pnt++;
-    if (*pnt) pnt++;
-    while(*pnt && *pnt != '~'){
-      prio1 = 10*prio1 + *pnt - '0';
-      pnt++;
-    };
-    priority = prio1;
-  };
-    
-  if (!dirflag){
-    if (!seen_dot && !omit_period) {
-      if (result) *result++ = '.'; 
-      extra++;
-    };
-    if(!omit_version_number && !seen_semic) {
-      if(result){
-	*result++ = ';';
-	*result++ = '1';
-      };
-      extra += 2;
-    }
-  };
-		    
-  if(result) *result++ = 0;
-  sresult->priority = priority;
-  return chars_before_dot + chars_after_dot + seen_dot + extra;
-}
-
-#ifdef ADD_FILES
-
-struct file_adds *root_file_adds = NULL;
-
-void
-FDECL2(add_one_file, char *, addpath, char *, path )
-{
-  char *cp;
-  char *name;
-  struct file_adds *f;
-  struct file_adds *tmp;
-
-  f = root_file_adds;
-  tmp = NULL;
-
-  name = rindex (addpath, PATH_SEPARATOR);
-  if (name == NULL) {
-    name = addpath;
-  } else {
-    name++;
-  }
-
-  cp = strtok (addpath, SPATH_SEPARATOR);
-
-  while (cp != NULL && strcmp (name, cp)) {
-     if (f == NULL) {
-        root_file_adds = e_malloc (sizeof *root_file_adds);
-        f=root_file_adds;
-        f->name = NULL;
-        f->child = NULL;
-        f->next = NULL;
-        f->add_count = 0;
-        f->adds = NULL;
-	f->used = 0;
-     }
-    if (f->child) {
-      for (tmp = f->child; tmp->next != NULL; tmp =tmp->next) {
-         if (strcmp (tmp->name, cp) == 0) {
-           f = tmp;
-           goto next;
-         }
-      }
-      if (strcmp (tmp->name, cp) == 0) {
-          f=tmp;
-          goto next;
-      }
-      /* add a new node. */
-      tmp->next = e_malloc (sizeof (*tmp->next));
-      f=tmp->next;
-      f->name = strdup (cp);
-      f->child = NULL;
-      f->next = NULL;
-      f->add_count = 0;
-      f->adds = NULL;
-      f->used = 0;
-    } else {
-      /* no children. */
-      f->child = e_malloc (sizeof (*f->child));
-      f = f->child;
-      f->name = strdup (cp);
-      f->child = NULL;
-      f->next = NULL;
-      f->add_count = 0;
-      f->adds = NULL;
-      f->used = 0;
-
-    }
-   next:
-     cp = strtok (NULL, SPATH_SEPARATOR);
-   }
-  /* Now f if non-null points to where we should add things */
-  if (f == NULL) {
-     root_file_adds = e_malloc (sizeof *root_file_adds);
-     f=root_file_adds;
-     f->name = NULL;
-     f->child = NULL;
-     f->next = NULL;
-     f->add_count = 0;
-     f->adds = NULL;
-   }
-
-  /* Now f really points to where we should add this name. */
-  f->add_count++;
-  f->adds = realloc (f->adds, sizeof (*f->adds)*f->add_count);
-  f->adds[f->add_count-1].path = strdup (path);
-  f->adds[f->add_count-1].name = strdup (name);
-}
-
-void
-FDECL3(add_file_list, int, argc, char **,argv, int, ind)
-{
-  char *ptr;
-  char *dup_arg;
-
-  while (ind < argc) {
-     dup_arg = strdup (argv[ind]);
-     ptr = index (dup_arg,'=');
-     if (ptr == NULL) {
-        free (dup_arg);
-        return;
-     }
-     *ptr = 0;
-     ptr++;
-     add_one_file (dup_arg, ptr);
-     free (dup_arg);
-     ind++;
-  }
-}
-void
-FDECL1(add_file, char *, filename)
-{
-  char buff[1024];
-  FILE *f;
-  char *ptr;
-  char *p2;
-  int count=0;
-
-  if (strcmp (filename, "-") == 0) {
-    f = stdin;
-  } else {
-    f = fopen (filename, "r");
-    if (f == NULL) {
-      perror ("fopen");
-      exit (1);
-    }
-  }
-  while (fgets (buff, 1024, f)) {
-    count++;
-    ptr = buff;
-    while (isspace (*ptr)) ptr++;
-    if (*ptr==0) continue;
-    if (*ptr=='#') continue;
-
-    if (ptr[strlen(ptr)-1]== '\n') ptr[strlen(ptr)-1]=0;
-    p2 = index (ptr, '=');
-    if (p2 == NULL) {
-      fprintf (stderr, "Error in line %d: %s\n", count, buff);
-      exit (1);
-    }
-    *p2 = 0;
-    p2++;
-    add_one_file (ptr, p2);
-  }
-  if (f != stdin) fclose (f);
-}
-
-#endif
+extern char * cdwrite_data;
 
 int FDECL2(main, int, argc, char **, argv){
   char * outfile;
   struct directory_entry de;
-  unsigned int mem_start;
+  unsigned long mem_start;
   struct stat statbuf;
   char * scan_tree;
+  char * merge_image = NULL;
+  struct iso_directory_record * mrootp = NULL;
   int c;
 #ifdef ADD_FILES
   char *add_file_file = NULL;
@@ -559,22 +343,34 @@ int FDECL2(main, int, argc, char **, argv){
   read_rcfile(argv[0]);
 
   outfile = NULL;
-  while ((c = getopt(argc, argv, "i:o:V:RrfvaTp:P:x:dDlLNzA:")) != EOF)
+  while ((c = getopt(argc, argv, "i:o:V:RrfvaTp:P:b:c:x:dDlLNzA:M:m:C:")) != EOF)
     switch (c)
       {
-      case 'p':
-	preparer = optarg;
-	if(strlen(preparer) > 128) {
-		fprintf(stderr,"Preparer string too long\n");
-		exit(1);
-	};
+      case 'C':
+	/*
+	 * This is a temporary hack until cdwrite gets the proper hooks in
+	 * it.
+	 */
+	cdwrite_data = optarg;
 	break;
-      case 'P':
-	publisher = optarg;
-	if(strlen(publisher) > 128) {
-		fprintf(stderr,"Publisher string too long\n");
+      case 'a':
+	all_files++;
+	break;
+      case 'b':
+	use_eltorito++;
+	boot_image = optarg;  /* pathname of the boot image on cd */
+	if (boot_image == NULL) {
+	        fprintf(stderr,"Required boot image pathname missing\n");
 		exit(1);
-	};
+	}
+	break;
+      case 'c':
+	use_eltorito++;
+	boot_catalog = optarg;  /* pathname of the boot image on cd */
+	if (boot_catalog == NULL) {
+	        fprintf(stderr,"Required boot catalog pathname missing\n");
+		exit(1);
+	}
 	break;
       case 'A':
 	appid = optarg;
@@ -589,20 +385,45 @@ int FDECL2(main, int, argc, char **, argv){
       case 'D':
 	RR_relocation_depth = 32767;
 	break;
+      case 'f':
+	follow_links++;
+	break;
+      case 'i':
+#ifdef ADD_FILES
+	add_file_file = optarg;
+	break;
+#else
+	usage();
+	exit(1);
+#endif
       case 'l':
 	full_iso9660_filenames++;
 	break;
       case 'L':
         allow_leading_dots++;
         break;
+      case 'M':
+	merge_image = optarg;
+	break;
       case 'N':
 	omit_version_number++;
 	break;
       case 'o':
 	outfile = optarg;
 	break;
-      case 'f':
-	follow_links++;
+      case 'p':
+	preparer = optarg;
+	if(strlen(preparer) > 128) {
+		fprintf(stderr,"Preparer string too long\n");
+		exit(1);
+	};
+	break;
+      case 'P':
+	publisher = optarg;
+	if(strlen(publisher) > 128) {
+		fprintf(stderr,"Publisher string too long\n");
+		exit(1);
+	};
 	break;
       case 'R':
 	use_RockRidge++;
@@ -611,18 +432,15 @@ int FDECL2(main, int, argc, char **, argv){
 	rationalize++;
 	use_RockRidge++;
 	break;
+      case 'T':
+	generate_tables++;
+	break;
       case 'V':
 	volume_id = optarg;
 	break;
       case 'v':
 	verbose++;
 	break;
-      case 'a':
-	all_files++;
-	break;
-      case 'T':
-	generate_tables++;
-	break;
       case 'z':
 #ifdef VMS
 	fprintf(stderr,"Transparent compression not supported with VMS\n");
@@ -631,14 +449,12 @@ int FDECL2(main, int, argc, char **, argv){
 	transparent_compression++;
 #endif
 	break;
+      case 'm':
+        add_match(optarg);
+	break;
       case 'x':
         exclude(optarg);
 	break;
-      case 'i':
-#ifdef ADD_FILES
-	add_file_file = optarg;
-	break;
-#endif
       default:
 	usage();
 	exit(1);
@@ -656,12 +472,18 @@ int FDECL2(main, int, argc, char **, argv){
 		}
 	}
 #endif
-  mem_start = (unsigned int) sbrk(0);
+#ifdef HAVE_SBRK
+  mem_start = (unsigned long) sbrk(0);
+#endif
 
   if(verbose) fprintf(stderr,"%s\n", version_string);
-  /* Now find the timezone offset */
 
-  timezone_offset = get_iso9660_timezone_offset();
+  if(     (cdwrite_data != NULL && merge_image == NULL)
+       || (cdwrite_data == NULL && merge_image != NULL) )
+    {
+      fprintf(stderr,"Multisession usage bug - both -C and -M must be specified.\n");
+      exit(0);
+    }
 
   /*  The first step is to scan the directory tree, and take some notes */
 
@@ -697,19 +519,65 @@ int FDECL2(main, int, argc, char **, argv){
 				       "THE IEEE P1282 PROTOCOL PROVIDES SUPPORT FOR POSIX FILE SYSTEM SEMANTICS",
 				       "PLEASE CONTACT THE IEEE STANDARDS DEPARTMENT, PISCATAWAY, NJ, USA FOR THE P1282 SPECIFICATION.", &extension_record_size);
 #endif
-  };
+  }
 
+  /*
+   * See if boot catalog file exists in root directory, if not
+   * we will create it.
+   */
+  if (use_eltorito)
+    init_boot_catalog(argv[optind]);
+
+  /*
+   * Find the device and inode number of the root directory.
+   * Record this in the hash table so we don't scan it more than
+   * once.
+   */
   stat_filter(argv[optind], &statbuf);
   add_directory_hash(statbuf.st_dev, STAT_INODE(statbuf));
 
+  memset(&de, 0, sizeof(de));
+
   de.filedir = root;  /* We need this to bootstrap */
-  scan_directory_tree(argv[optind], &de);
+
+  if( merge_image != NULL )
+    {
+      mrootp = merge_isofs(merge_image);
+      if( mrootp == NULL )
+	{
+	  /*
+	   * Complain and die.
+	   */
+	  fprintf(stderr,"Unable to open previous session image %s\n",
+		  merge_image);
+	  exit(1);
+	}
+
+      memcpy(&de.isorec.extent, mrootp->extent, 8);      
+    }
+
+  /*
+   * Scan the actual directory (and any we find below it)
+   * for files to write out to the output image.
+   */
+  if (!scan_directory_tree(argv[optind], &de, mrootp))
+    {
+      exit(1);
+    }
+
+  /*
+   * Fix a couple of things in the root directory so that everything
+   * is self consistent.
+   */
   root->self = root->contents;  /* Fix this up so that the path tables get done right */
 
   if(reloc_dir) sort_n_finish(reloc_dir);
 
   if (goof) exit(1);
   
+  /*
+   * OK, ready to write the file.  Open it up, and generate the thing.
+   */
   if (outfile){
 	  discimage = fopen(outfile, "w");
 	  if (!discimage){
@@ -724,12 +592,14 @@ int FDECL2(main, int, argc, char **, argv){
 
   path_blocks = (path_table_size + (SECTOR_SIZE - 1)) >> 11;
   if (path_blocks & 1) path_blocks++;
-  path_table[0] = 0x14;
-  path_table[1] = path_table[0] + path_blocks;
-  path_table[2] = path_table[1] + path_blocks;
-  path_table[3] = path_table[2] + path_blocks;
 
-  last_extent = path_table[3] + path_blocks;  /* The next free block */
+  path_table[0] = session_start + 0x14;
+  path_table[1] = 0;
+  path_table[2] = path_table[0] + path_blocks;
+  path_table[3] = 0;
+
+  last_extent += path_table[2] - session_start + path_blocks;  
+		/* The next free block */
 
   /* The next step is to go through the directory tree and assign extent
      numbers for all of the directories */
@@ -740,9 +610,9 @@ int FDECL2(main, int, argc, char **, argv){
 	  struct directory_entry * s_entry;
 	  extension_record_extent = last_extent++;
 	  s_entry = root->contents;
-	  set_733(s_entry->rr_attributes + s_entry->rr_attr_size - 24,
+	  set_733((char *) s_entry->rr_attributes + s_entry->rr_attr_size - 24,
 		  extension_record_extent);
-	  set_733(s_entry->rr_attributes + s_entry->rr_attr_size - 8,
+	  set_733((char *) s_entry->rr_attributes + s_entry->rr_attr_size - 8,
 		  extension_record_size);
   };
 
@@ -759,10 +629,17 @@ int FDECL2(main, int, argc, char **, argv){
   if (verbose)
     dump_tree(root);
 
+  if( in_image != NULL )
+    {
+      fclose(in_image);
+    }
+
   iso_write(discimage);
 
+#ifdef HAVE_SBRK
   fprintf(stderr,"Max brk space used %x\n", 
-	  ((unsigned int)sbrk(0)) - mem_start);
+	  (unsigned int)(((unsigned long)sbrk(0)) - mem_start));
+#endif
   fprintf(stderr,"%d extents written (%d Mb)\n", last_extent, last_extent >> 9);
 #ifdef VMS
   return 1;
@@ -771,11 +648,12 @@ int FDECL2(main, int, argc, char **, argv){
 #endif
 }
 
-void *e_malloc(size_t size)
+void *
+FDECL1(e_malloc, size_t, size)
 {
-void* pt;
-	if((pt=malloc(size))==NULL) {
-		printf("Not enougth memory\n");
+void* pt = 0;
+	if( (size > 0) && ((pt=malloc(size))==NULL) ) {
+		fprintf(stderr, "Not enough memory\n");
 		exit (1);
 		}
 return pt;
