@@ -19,7 +19,7 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
-static char rcsid[] ="$Id: write.c,v 1.3 1997/03/08 16:24:02 eric Rel $";
+static char rcsid[] ="$Id: write.c,v 1.4 1997/04/10 03:33:25 eric Rel $";
 
 #include <string.h>
 #include <stdlib.h>
@@ -305,6 +305,17 @@ int FDECL2(compare_dirs, const void *, rr, const void *, ll)
      rpnt = (*r)->isorec.name;
      lpnt = (*l)->isorec.name;
      
+     /*
+      *  Put the '.' and '..' entries on the head of the sorted list.
+      *  For normal ASCII, this always happens to be the case, but out of
+      *  band characters cause this not to be the case sometimes.
+      */
+     if( strcmp(rpnt, ".") == 0 ) return -1;
+     if( strcmp(lpnt, ".") == 0 ) return  1;
+
+     if( strcmp(rpnt, "..") == 0 ) return -1;
+     if( strcmp(lpnt, "..") == 0 ) return  1;
+
      while(*rpnt && *lpnt) 
      {
 	  if(*rpnt == ';' && *lpnt != ';') return -1;
